@@ -4,7 +4,8 @@
 from __future__ import absolute_import, division, print_function  #, unicode_literals
 # __package__ = 'cma'
 import warnings as _warnings
-import collections as _collections, functools as _functools
+import collections as _collections
+import functools as _functools
 import numpy as np
 from numpy import logical_and as _and, logical_or as _or, logical_not as _not
 from .utilities.utils import rglen, is_
@@ -15,6 +16,8 @@ from .optimization_tools import BestSolution2
 from .utilities.python3for2 import range
 del absolute_import, division, print_function  #, unicode_literals
 
+_warnings.filterwarnings('once', message="``import moarchiving`` failed.*")
+
 class BoundaryHandlerBase(object):
     """quick hack versatile base class"""
     def __init__(self, bounds):
@@ -51,6 +54,41 @@ class BoundaryHandlerBase(object):
                                      'therefore no finite feasible solution is available')
             self.bounds = bounds
 
+    def amend_bounds_for_integer_variables(self, integer_indices, at=0.5, offset=1e-9):
+        """set bounds away from ``at=0.5`` such that
+
+        a repaired solution is always rounded into the feasible domain.
+        """
+        if not integer_indices or not self.has_bounds():
+            return
+        def set_bounds(which, idx, integer_indices):
+            """idx is a `bool` index array of ``bound % 1 == 0.5``"""
+            assert which in (0, 1), which
+            if not np.any(idx):
+                return
+            for i in np.nonzero(idx)[0]:
+                if i not in integer_indices:
+                    idx[i] = False
+            if not np.any(idx):
+                return
+            dimension = max((len(self.bounds[which]),
+                             np.max(np.nonzero(idx)[0]) + 1))
+            bounds = self.get_bounds(which, dimension)
+            if len(bounds) < len(idx):
+                idx = idx[:len(bounds)]  # last nonzero entry determined len
+            elif len(bounds) > len(idx):
+                idx = np.hstack([idx, np.zeros(len(bounds) - len(idx), dtype=bool)])
+            bounds[idx] += (1 - 2 * which) * offset * np.maximum(1, np.abs(bounds[idx]))
+            self.bounds[which] = bounds
+            self._bounds_dict = {}
+
+        dimension = max(integer_indices) + 1
+        for which in (0, 1):
+            if not self.has_bounds('upper' if which else 'lower'):
+                continue
+            bounds = self.get_bounds(which, dimension)
+            set_bounds(which, np.mod(bounds, 1) == at, integer_indices)
+
     def __call__(self, solutions, *args, **kwargs):
         """return penalty or list of penalties, by default zero(s).
 
@@ -106,25 +144,33 @@ class BoundaryHandlerBase(object):
         sign_ = 2 * ib - 1
         assert sign_**2 == 1
         if self.bounds is None or self.bounds[ib] is None:
-            return np.array(dimension * [sign_ * np.Inf])
+            return np.array(dimension * [sign_ * np.inf])
         res = []
         for i in range(dimension):
             res.append(self.bounds[ib][min([i, len(self.bounds[ib]) - 1])])
             if res[-1] is None:
-                res[-1] = sign_ * np.Inf
+                res[-1] = sign_ * np.inf
         return np.array(res)
 
-    def has_bounds(self):
+    def has_bounds(self, which='both'):
         """return `True` if any variable is bounded"""
+        valid_whichs = (None, 'both', 'lower', 'upper')
+        if which not in valid_whichs:
+            raise ValueError("`which` parameter must be in {0} but was ={1}"
+                                .format(valid_whichs, which))
         bounds = self.bounds
-        if bounds is None or all(b is None for b in bounds):
+        if bounds is None or bounds in (False, [], ()) or (
+                bounds[0] is None and bounds[1] is None):
             return False
         for ib, bound in enumerate(bounds):
-            if bound is not None:
-                sign_ = 2 * ib - 1
-                for bound_i in bound:
-                    if bound_i is not None and sign_ * bound_i < np.inf:
-                        return True
+            if bound is None or (
+                ib == 0 and which == 'upper') or (
+                ib == 1 and which == 'lower'):
+                continue
+            sign_ = 2 * ib - 1
+            for bound_i in bound:
+                if bound_i is not None and sign_ * bound_i < np.inf:
+                    return True
         return False
 
     def is_in_bounds(self, x):
@@ -164,6 +210,19 @@ class BoundaryHandlerBase(object):
                    idxs += [i]
         return sorted(idxs)
 
+    def get_bound(self, index):
+        """return lower and upper bound of variable with index `index`"""
+        if self.bounds is None:
+            return [-np.inf, np.inf]
+        res = []
+        for ib in [0, 1]:
+            if self.bounds[ib] is None or len(self.bounds[ib]) == 0:
+                b = None
+            else:
+                b = self.bounds[ib][min((index, len(self.bounds[ib]) - 1))]
+            res.append([-np.inf, np.inf][ib] if b is None else b)
+        return res
+
     def to_dim_times_two(self, bounds):
         """return boundaries in format ``[[lb0, ub0], [lb1, ub1], ...]``,
         as used by ``BoxConstraints...`` class.
@@ -217,11 +276,11 @@ class BoundTransform(BoundaryHandlerBase):
     >>> from cma.constraints_handler import BoundTransform
     >>> from cma import fitness_transformations as ft
     >>> veq = cma.utilities.math.Mh.vequals_approximately
-    >>> b = BoundTransform([None, 1])
-    >>> assert b.bounds == [[None], [1]]
-    >>> assert veq(b.repair([0, 1, 1.2]), np.array([ 0., 0.975, 0.975]))
+    >>> b = BoundTransform([0, None])
+    >>> assert b.bounds == [[0], [None]]
+    >>> assert veq(b.repair([-0.1, 0, 1, 1.2]), np.array([0.0125, 0.0125, 1, 1.2])), b.repair([-0.1, 0, 1, 1.2])
     >>> assert b.is_in_bounds([0, 0.5, 1])
-    >>> assert veq(b.transform([0, 1, 2]), [ 0.   ,  0.975,  0.2  ])
+    >>> assert veq(b.transform([-1, 0, 1, 2]), [0.9, 0.0125,  1,  2  ]), b.transform([-1, 0, 1, 2])
     >>> bounded_sphere = ft.ComposedFunction([
     ...         cma.ff.sphere,
     ...         BoundTransform([[], 5 * [-1] + [np.inf]]).transform
@@ -244,6 +303,13 @@ class BoundTransform(BoundaryHandlerBase):
     ...     assert veq(x_to_b, x2_to_b)
     ...     assert veq(x2, x3)
     ...     assert veq(x2_to_b, x3_to_b)
+    >>> for _ in range(5):
+    ...     lb = np.random.randn(4)
+    ...     ub = lb + 1e-7 + np.random.rand(4)
+    ...     b = BoundTransform([lb, ub])
+    ...     for x in [np.random.randn(4) / np.sqrt(np.random.rand(4)) for _ in range(22)]:
+    ...         assert all(lb <= b.transform(x)), (lb, ub, b.__dict__)
+    ...         assert all(b.transform(x) <= ub), (lb, ub, b.__dict__)
 
     Details: this class uses ``class BoxConstraintsLinQuadTransformation``
 
@@ -295,7 +361,7 @@ class BoundPenalty(BoundaryHandlerBase):
     options (`CMAOptions`) of the class `cma.CMAEvolutionStrategy`:
 
     >>> import cma
-    >>> res = cma.fmin(cma.ff.elli, 6 * [1], 1,
+    >>> res = cma.fmin(cma.ff.elli, 6 * [0.9], 0.1,
     ...     {'BoundaryHandler': cma.BoundPenalty,
     ...      'bounds': [-1, 1],
     ...      'tolflatfitness': 10,
@@ -306,7 +372,8 @@ class BoundPenalty(BoundaryHandlerBase):
 
     Reference: Hansen et al 2009, A Method for Handling Uncertainty...
     IEEE TEC, with addendum, see
-    http://www.lri.fr/~hansen/TEC2009online.pdf
+    https://ieeexplore.ieee.org/abstract/document/4634579
+    https://hal.inria.fr/inria-00276216/file/TEC2008.pdf
 
     **todo**: implement a more generic interface, where this becomes a
     fitness wrapper which adds the desired penalty and the `update`
@@ -342,8 +409,8 @@ class BoundPenalty(BoundaryHandlerBase):
         # TODO (old data): CPU(N,lam,iter=20,200,100): 3.3s of 8s for two bounds, 1.8s of 6.5s for one bound
         # remark: np.max([bounds[0], x]) is about 40 times slower than max((bounds[0], x))
         copy = copy_if_changed
-        bounds = self.bounds
-        if bounds not in (None, [None, None], (None, None)):  # solely for effiency
+        if self.has_bounds():
+            bounds = self.bounds
             if copy:
                 x = np.array(x, copy=True)
             if bounds[0] is not None:
@@ -375,7 +442,7 @@ class BoundPenalty(BoundaryHandlerBase):
         """
         # if x in (None, (), []):  # breaks when x is a nparray
         #     return x
-        if self.bounds in (None, [None, None], (None, None)):
+        if not self.has_bounds():
             return 0.0 if np.isscalar(x[0]) else [0.0] * len(x)  # no penalty
 
         x_is_single_vector = np.isscalar(x[0])
@@ -387,7 +454,7 @@ class BoundPenalty(BoundaryHandlerBase):
             gamma = list(self.gamma)  # fails if self.gamma is a scalar
             for i in sorted(gp.fixed_values):  # fails if fixed_values is None
                 gamma.insert(i, 0.0)
-            gamma = np.array(gamma, copy=False)
+            gamma = np.asarray(gamma)
         except TypeError:
             gamma = self.gamma
         pen = []
@@ -395,7 +462,7 @@ class BoundPenalty(BoundaryHandlerBase):
             # CAVE: this does not work with already repaired values!!
             # CPU(N,lam,iter=20,200,100)?: 3s of 10s, np.array(xi): 1s
             # remark: one deep copy can be prevented by xold = xi first
-            xpheno = gp.pheno(archive[xi]['geno'])
+            xpheno = xi if gp.isidentity else gp.pheno(archive[xi]['geno'])
             # necessary, because xi was repaired to be in bounds
             xinbounds = self.repair(xpheno)
             # could be omitted (with unpredictable effect in case of external repair)
@@ -446,8 +513,8 @@ class BoundPenalty(BoundaryHandlerBase):
 
         # ## Store/update a history of delta fitness value
         fvals = sorted(function_values)
-        l = 1 + len(fvals)
-        val = fvals[3 * l // 4] - fvals[l // 4]  # exact interquartile range apart interpolation
+        _l = 1 + len(fvals)
+        val = fvals[3 * _l // 4] - fvals[_l // 4]  # exact interquartile range apart interpolation
         val = val / np.mean(varis)  # new: val is normalized with sigma of the same iteration
         # insert val in history
         if np.isfinite(val) and val > 0:
@@ -505,7 +572,7 @@ def _g_pos_sum(gvals):
     return sum(gi for gi in gvals if gi > 0)
 def _g_pos_squared_sum(gvals):
     return sum(gi**2 for gi in gvals if gi > 0)
-class ConstrainedSolutionsArchive:
+class ConstrainedSolutionsArchive(object):
     """Biobjective Pareto archive to store some Pareto optimal solutions
     for constrained optimization.
 
@@ -525,7 +592,6 @@ class ConstrainedSolutionsArchive:
         except ImportError:
             m = ("``import moarchiving`` failed, hence convergence tracking "
                  "is disabled. \n  'pip install moarchiving' should fix this.")
-            _warnings.filterwarnings('once', message=m)
             _warnings.warn(m)
             return
         self.archive = moarchiving.BiobjectiveNondominatedSortedList()
@@ -533,8 +599,10 @@ class ConstrainedSolutionsArchive:
         self.count += 1
         if self.archive is not None:
             gagg = self.aggregator(g)
-            try: self.archive.add([f, gagg], info=info)
-            except TypeError: self.archive.add([f, gagg])  # previous interface
+            try:
+                self.archive.add([f, gagg], info=info)
+            except TypeError:
+                self.archive.add([f, gagg])  # previous interface
             while len(self.archive) > self.maxlen:
                 if self.archive[1][1] > 0:  # keep at least one infeasible solution
                     self.archive.remove(self.archive[0])
@@ -623,8 +691,9 @@ class DequeCDF(_collections.deque):
     >>> d = cma.constraints_handler.DequeCDF(maxlen=22)
     >>> for i in range(5):
     ...     d.append([i])
-    >>> d.cdf(0, 0), d.cdf(0, 2), d.cdf(0, 2.1), d.cdf(0, 22.1), d.cdf(0, 4, 2)
-    (0.1, 0.5, 0.6, 1.0, 0.75)
+    >>> ' '.join(['{0:.2}'.format(x) for x in
+    ...                   [d.cdf(0, 0), d.cdf(0, 2), d.cdf(0, 2.1), d.cdf(0, 22.1), d.cdf(0, 4, 2)]])
+    '0.1 0.5 0.6 1.0 0.75'
 
     """
     def cdf(self, i, val=0, len_=None):
@@ -686,7 +755,7 @@ def _log_mu(s):
     v = np.log10(np.maximum(s.mu, 1e-9))
     return np.hstack([v, 0])  # add column to get same colors as _log_feas_events
 
-class CountLastSameChanges:
+class CountLastSameChanges(object):
     """An array/list of successive same-sign counts.
 
     ``.same_changes[i]`` counts how often the same sign was successively
@@ -713,7 +782,7 @@ class CountLastSameChanges:
         if force or self.chi_exponent_threshold is None:  # threshold for number of consecutive changes
             # self.chi_exponent_threshold = 5 + self.dimension**0.5  # previous value, probably too small in >- 20-D
             # self.chi_exponent_threshold = 5 + self.dimension**0.75  # sweeps in aug-lag4-mu-update.ipynv
-            self.chi_exponent_threshold = 2 + self.dimension 
+            self.chi_exponent_threshold = 2 + self.dimension
         if force or self.chi_exponent_factor is None:  # factor used in shape_exponent
             # self.chi_exponent_factor = 3 / dimension**0.5 # previous value
             # self.chi_exponent_factor = 1 * dimension**0.25 / self.chi_exponent_threshold  # steepness of ramp up
@@ -811,7 +880,7 @@ class AugmentedLagrangian(object):
     More testing based on the simpler `ConstrainedFitnessAL` interface:
 
     >>> import cma
-    >>> for algorithm, evals in zip((0, 1, 2, 3), (2000, 2200, 1500, 1800)): 
+    >>> for algorithm, evals in zip((0, 1, 2, 3), (2000, 2200, 1500, 1800)):
     ...     alf = cma.ConstrainedFitnessAL(cma.ff.sphere, lambda x: [x[0] + 1], 3,
     ...                                    find_feasible_first=True)
     ...     _ = alf.al.set_algorithm(algorithm)
@@ -872,7 +941,7 @@ class AugmentedLagrangian(object):
         elif self.algorithm == 2:
             self.set_dufosse2020()
         else:
-            raise ValueError("Algorithm id {} is not recognized".format(self.algorithm))
+            raise ValueError("Algorithm id {0} is not recognized".format(self.algorithm))
     def set_algorithm(self, algorithm=None):
         """if algorithm not `None`, set it and return self,
 
@@ -917,8 +986,10 @@ class AugmentedLagrangian(object):
         return len(self.lam)
     @property
     def is_initialized(self):
-        try: return all(self._initialized)
-        except TypeError: return bool(self._initialized)
+        try:
+            return all(self._initialized)
+        except TypeError:
+            return bool(self._initialized)
     @property
     def count_initialized(self):
         """number of constraints with initialized coefficients"""
@@ -926,8 +997,10 @@ class AugmentedLagrangian(object):
     @property
     def feasibility_ratios(self):
         """or bias for equality constraints, versatile interface may change"""
-        try: return [np.mean(np.asarray(g) <= 0) for g in np.asarray(self.G).T]
-        except AttributeError: return None
+        try:
+            return [np.mean(np.asarray(g) <= 0) for g in np.asarray(self.G).T]
+        except AttributeError:
+            return None
 
     def set_m(self, m):
         """initialize attributes which depend on the number of constraints.
@@ -952,7 +1025,7 @@ class AugmentedLagrangian(object):
 
         The formulas to set the coefficients::
 
-            lam = iqr(f) / (sqrt(n) * iqr(g))
+            lam = iqr(f) / (n * iqr(g))
             mu = 2 * iqr(f) / (5 * n * (iqr(g) + iqr(g**2)))
 
         are taken out of thin air and not thoroughly tested. They are
@@ -1036,8 +1109,10 @@ class AugmentedLagrangian(object):
         Penalties are zero in the optimum and can be negative down to
         ``-lam**2 / mu / 2``.
         """
-        try: self.count_g_in_penalized_domain += self.mu * g > -1 * self.lam
-        except TypeError: pass
+        try:
+            self.count_g_in_penalized_domain += self.mu * g > -1 * self.lam
+        except TypeError:
+            pass
         for i in range(len(g)):
             self.g_all[i].append(g[i])
         if self.lam is None:
@@ -1122,8 +1197,10 @@ class AugmentedLagrangian(object):
         """
         self.g_history.append(g)  # is a deque with maxlen
         if self.lam is None:
-            try: self._count_noupdate += 1
-            except AttributeError: self._count_noupdate = 1
+            try:
+                self._count_noupdate += 1
+            except AttributeError:
+                self._count_noupdate = 1
             if self._count_noupdate % (1 + self._count_noupdate**0.5) < 1:
                 _warnings.warn("no update for %d calls (`lam` and `mu` need "
                                "to be initialized first)" % self._count_noupdate)
@@ -1195,7 +1272,8 @@ class AugmentedLagrangian(object):
                         self.mu[i] *= self.chi_domega**(-0.25 * (1 -
                             self.count_mu_last_same_changes.update(i, -1).shape_exponent(i)))
                 else:
-                    raise NotImplemented("algorithm number {} is not known".format(self.algorithm))
+                    raise NotImplementedError("algorithm number {0} is not known"
+                                              .format(self.algorithm))
             self.count += 1
         assert np.all((self.lam >= 0) + self.isequality)
         self.f, self.g = f, g  # self(g) == 0 if mu=lam=0
@@ -1208,7 +1286,7 @@ class AugmentedLagrangian(object):
 def _get_favorite_solution(es):
     "avoid unpicklable lambda construct"
     return es.ask(1, sigma_fac=0)[0]
-class ConstrainedFitnessAL:
+class ConstrainedFitnessAL(object):
     """Construct an unconstrained objective function from constraints.
 
     This class constructs an unconstrained "fitness" function (to be
@@ -1260,6 +1338,8 @@ class ConstrainedFitnessAL:
     >>> x = cfun.find_feasible(es)  # uses es.optimize to find (another) feasible solution
     >>> assert constraints(x)[0] <= 0, (x, cfun.best_feas.x)
     >>> assert cfun.best_feas.f < 1 + 2e-6, str(cfun.best_feas)
+    >>> assert len(cfun.archives) == 3
+    >>> assert cma.ConstrainedFitnessAL(cma.ff.sphere, constraints, archives=False).archives == []
 
     Details: The fitness, to be minimized, is changing over time such that
     the overall minimal value does not indicate the best solution.
@@ -1276,7 +1356,7 @@ class ConstrainedFitnessAL:
     (g_i > 0) x g_i^2`` and omits ``f + sum_i lam_i g_i`` altogether.
     Whenever a feasible solution is found, the `finding_feasible` flag is
     reset to `False`.
-    
+
     `find_feasible(es)` sets ``finding_feasible = True`` and uses
     `es.optimize` to optimize `self.__call__`. This works well with
     `CMAEvolutionStrategy` but may easily fail with solvers that do not
@@ -1337,7 +1417,7 @@ class ConstrainedFitnessAL:
         try:  # treat archives as list of constraints aggregation functions
             self.archives = [ConstrainedSolutionsArchive(fun) for fun in archives]
         except TypeError:  # treat archives as flag
-            if self.archives:
+            if archives:
                 self.archives = [ConstrainedSolutionsArchive(fun) for fun in
                                  ConstrainedFitnessAL.archive_aggregators]
             else:
@@ -1432,17 +1512,18 @@ class ConstrainedFitnessAL:
                 self.find_feasible_aggregator, aggregator_ = aggregator, self.find_feasible_aggregator
             while self.finding_feasible and not any(any(d == m for m in termination)
                                                     for d in es.stop()):
-                es.optimize(self, 1)  # we could put the below fcts as callbacks
+                X = es.ask()
+                es.tell(X, [self(x) for x in X])
                 self.update(es)  # calls fun if finding_feasible finished and method is not 'best'
+                es.logger.add()  # relies on calling log_in_es during update
                 if self.finding_feasible:
                     self._reset_arrays()
-                es.logger.add()
             if aggregator:  # reset `find_feasible_aggregator` to original value
                 self.find_feasible_aggregator = aggregator_
         # warn when no feasible solution was found
         if self.best_feas.x is None or self.finding_feasible:
             _warnings.warn("ConstrainedFitnessAL.find_feasible: "
-                           " No {}feasible solution found, stop() == {}".format(
+                           " No {0}feasible solution found, stop() == {1}".format(
                                "new " if self.best_feas.x is not None else "", es.stop()))
         assert self.best_feas.x is not None or self.finding_feasible, (self.best_feas, self.finding_feasible)
         return self.best_feas.x
@@ -1483,7 +1564,7 @@ class ConstrainedFitnessAL:
               When method is not best, it should work without even call self(xfavorite).
         """
         if not len(self.F) == len(self.G) == len(self.F_plus_sum_al_G):
-            _warnings.warn("len(F, G, F_plus_sum_al_G) = ({}, {}, {}) differ."
+            _warnings.warn("len(F, G, F_plus_sum_al_G) = ({0}, {1}, {2}) differ."
                            "This is probably a bug!".format(
                            len(self.F), len(self.G), len(self.F_plus_sum_al_G)))
         if len(self.G) == 0:  # TODO: we could first self(es.result.xfavorite) and should be fine
@@ -1547,6 +1628,5 @@ class ConstrainedFitnessAL:
             es.more_to_write += [g_pos if g_pos else np.nan,  # nan avoids zero in log plot
                                  10**(n_infeas / 10) if n_infeas < 10 else n_infeas,  # symlog-like
                                  ]
-        except:
+        except Exception:
             pass
-
