@@ -189,9 +189,9 @@ class EvalParallel2(object):
     """A class and context manager for parallel evaluations.
 
     This class is based on the ``Pool`` class of the `multiprocessing` module.
-    
+
     The interface in v2 changed, such that the fitness function can be
-    given once in the constructor. Hence the number of processes has 
+    given once in the constructor. Hence the number of processes has
     become the second (optional) argument of `__init__` and the function
     has become the second and optional argument of `__call__`.
 
@@ -225,7 +225,7 @@ class EvalParallel2(object):
     ...         res = eval_all([[1,2], [3,4]])
     >>> # class usage, don't forget to call terminate
     >>> ep = EvalParallel2(cma.fitness_functions.elli, 4)
-    >>> ep([[1,2], [3,4], [4, 5]])  # doctest:+ELLIPSIS
+    >>> [float(v) for v in ep([[1,2], [3,4], [4, 5]])]  # doctest:+ELLIPSIS
     [4000000.944...
     >>> ep.terminate()
     ...
@@ -481,7 +481,7 @@ class BinaryEvolutionPath(EvolutionPath):
     def raw_binary_s(self):
         """return one of two possible values with expectation of zero.
 
-        the maximum for the larger value is 1 - 0.15865525393145707 for tc\to\infty.
+        the maximum for the larger value is 1 - 0.15865525393145707 for tc to infty.
         """
         # p * (I - p) + (1 - p) * (0 - p) = p - p^2  - p + p^2 = 0
         return (np.abs(self.values) > 1) - self.probability_larger_than_one_from_binary
@@ -793,17 +793,17 @@ class NoiseHandler(object):
         if choice == 1:
             # take n_first first and reev - n_first best of the remaining
             n_first = lam_reev - lam_reev // 2
-            sort_idx = np.argsort(np.array(fit, copy=False)[n_first:]) + n_first
-            return np.array(list(range(0, n_first)) +
-                            list(sort_idx[0:lam_reev - n_first]), copy=False)
+            sort_idx = np.argsort(np.asarray(fit)[n_first:]) + n_first
+            return np.asarray(list(range(0, n_first)) +
+                            list(sort_idx[0:lam_reev - n_first]))
         elif choice == 2:
-            idx_sorted = np.argsort(np.array(fit, copy=False))
+            idx_sorted = np.argsort(np.asarray(fit))
             # take lam_reev equally spaced, starting with best
             linsp = np.linspace(0, len(fit) - len(fit) / lam_reev, lam_reev)
             return idx_sorted[[int(i) for i in linsp]]
         # take the ``lam_reeval`` best from the first ``2 * lam_reeval + 2`` values.
         elif choice == 3:
-            return np.argsort(np.array(fit, copy=False)[:2 * (lam_reev + 1)])[:lam_reev]
+            return np.argsort(np.asarray(fit)[:2 * (lam_reev + 1)])[:lam_reev]
         else:
             raise ValueError('unrecognized choice value %d for noise reev'
                              % choice)
@@ -887,7 +887,7 @@ class Sections(object):
         self.basis = np.eye(len(x)) if basis is None else basis
 
         try:
-            self.load()
+            load and self.load()
             if any(self.res['x'] != x):
                 self.res = {}
                 self.res['x'] = x  # TODO: res['x'] does not look perfect
@@ -999,4 +999,3 @@ class Sections(object):
         s = pickle.load(open(name + '.pkl', 'rb'))
         self.res = s.res  # disregard the class
         return self
-
